unit KnobsFiles;

{

   COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

// File reading and tokenizing


interface

uses

  Windows, SysUtils, Classes, SyncObjs;

const

  WhiteSpace : set of AnsiChar = [ #$00 .. #$20, #$7f .. #$ff];

  endoffile = ';s';    // special token denoting end of file

type

  TOnFileLine    = procedure( aSender: TObject; aLineNr, aCount: Integer)      of object;
  TOnLine        = procedure( aSender: TObject; aLineNr, aTotalCount: Integer) of object;
  TOnFileEnter   = procedure( aSender: TObject; const aFileName: string)       of object;
  TOnFileLeave   = procedure( aSender: TObject; const aFileName: string)       of object;
  TOnFileResume  = procedure( aSender: TObject; const aFileName: string)       of object;


  TTokenFile = class( TStringList)
  private
    FFileName     : string;
    FCurrToken    : string;
    FLastToken    : string;
    FBuffer       : string;
    FOldLineNr    : Integer;
    FLineNr       : Integer;
    FOffset       : Integer;
    FFbDecimation : Integer;
    FLineStack    : TStringList;
    FOnFileLine   : TOnFileLine;
  private
    procedure   Reset;                                                  virtual;
    procedure   FillBuffer;
    Function    GetOffset: Integer;
  public
    constructor Create( const aFileName: string);                       virtual;
    destructor  Destroy;                                               override;
    procedure   BreakOut;                                               virtual;
    function    GetString( aTerminator: Char): string;
    function    GetToken( aSeparator: Char; Extended: Boolean): string;
    procedure   SkipToEol;
    function    ReadLine: string;
    function    GetToEol: string;
    procedure   PushBackBuffer;
    procedure   PushLine( const aLine: string);                         virtual;
    function    PopLine: string;                                        virtual;
  public
    property    FileName     : string      read FFileName;
    property    LineNr       : Integer     read FLineNr;
    property    Offset       : Integer     read GetOffset;
    property    FbDecimation : Integer     read FFbDecimation write FFbDecimation;
    property    OnFileLine   : TOnFileLine read FOnFileLine   write FOnFileLine;
  end;

  TFileStack = class( TList)
  private
    FOnLine       : TOnLine;
    FOnFileEnter  : TOnFileEnter;
    FOnFileLeave  : TOnFileLeave;
    FOnFileResume : TOnFileResume;
    FTotalCount   : Integer;
    FFbDecimation : Integer;
  private
    function    GetTokenFile: TTokenFile;
    function    GetFileName : string;
    function    GetLineNr   : Integer;
    function    GetOffset   : Integer;
    procedure   SetFbDecimation( aValue: Integer);
    function    GetFileCount: Integer;
    function    GetFileIndexed( anIndex: Integer): TTokenFile;
  private
    procedure   CloseAll;
    procedure   Pop;
    procedure   PushBackBuffer;
    procedure   DoFileLine( aSender: TObject; aLineNr, aCount: Integer);
  public
    constructor Create;
    destructor  Destroy;                                               override;
    procedure   BreakOut;
    procedure   Reset;
    procedure   Push( const aFileName: string);
    function    GetToken( aSeparator: Char; Extended: Boolean; out aFile: string; out aLine, anOffset: Integer): string;
    procedure   SkipToEol;
    function    ReadLine: string;
    function    GetToEol: string;
    procedure   PushLine( const aLine: string; const anID: string);
    procedure   PushLines( const aLines: TStrings; const anID: string);
  private
    property    TokenFile      : TTokenFile    read GetTokenFile;
    property    FileCount      : Integer       read GetFileCount;
  private
    property    FileIndexed[ anIndex: Integer]: TTokenFile read GetFileIndexed;
  public
    property    FileName       : string        read GetFileName;
    property    LineNr         : Integer       read GetLineNr;
    property    Offset         : Integer       read GetOffset;
    property    FbDecimation   : Integer       read FFbDecimation  write SetFbDecimation;
  public
    property    OnLine         : TOnLine       read FOnLine        write FOnLine;
    property    OnFileEnter    : TOnFileEnter  read FOnFileEnter   write FOnFileEnter;
    property    OnFileLeave    : TOnFileLeave  read FOnFileLeave   write FOnFileLeave;
    property    OnFileResume   : TOnFileResume read FOnFileResume  write FOnFileResume;
  end;


    function  StrToUIntBase( const aValue: string; aBase: Byte): Cardinal;


implementation



    function  StrToUIntBase( const aValue: string; aBase: Byte): Cardinal;

      function LookUp( aDig: char): Byte;
      var
        n : Integer;
      begin
        Result := $ff;
        if ( aDig >= '0') and ( aDig <= 'Z')
        then begin
          n := Ord( aDig) - Ord( '0');
          if n > 9
          then n := 10 + n - ( Ord( 'A') - Ord( '0'));
          Result := n;
        end;
      end;

    var
      C : Char;
      N : Byte;
    begin
      Result := 0;
      if ( aBase >= 2) and ( aBase <= 36)
      then begin
        for C in aValue
        do begin
          N := Lookup( UpCase( C));
          if ( N < aBase) and ( N <> $ff)
          then Result := Result * aBase + N
          else Break;
        end;
      end;
    end;


{ ========
  TTokenFile = class( TObject)
  private
    FFileName     : string;
    FCurrToken    : string;
    FLastToken    : string;
    FBuffer       : string;
    FOldLineNr    : Integer;
    FLineNr       : Integer;
    FOffset       : Integer;
    FFbDecimation : Integer;
    FLineStack    : TStringList;
    FOnLine       : TOnFileLine;
  public
    property    FileName     : string      read FFileName;
    property    LineNr       : Integer     read FLineNr;
    property    Offset       : Integer     read GetOffset;
    property    FbDecimation : Integer     read FFbDecimation write FFbDecimation;
    property    OnLine       : TOnFileLine read FOnLine       write FOnLine;
  private
}


    procedure   TTokenFile.Reset; // virtual;
    begin
      FBuffer    := '';
      FCurrToken := '';
      FlastToken := '';
      FLineNr    := 0;
      FOffset    := 1;
      FLineStack.Clear;
    end;

    procedure   TTokenFile.FillBuffer;
    begin
      if FLineStack.Count > 0
      then FBuffer := PopLine
      else begin
        FBuffer := Strings[ FLineNr];
        Inc( FLineNr);
        if Assigned( FOnFileLine)
        then begin
          if FLineNr mod FFbDecimation = 0
          then begin
            FOnFileLine( Self, FLineNr, FLineNr - FOldLineNr);
            FOldLineNr := FLineNr;
          end;
        end;
        FOffset := 1;
      end;
    end;

    function    TTokenFile.GetOffset: Integer;
    begin
      Result := FOffset - Length( FLastToken);
      if Result <= 0
      then Result := 1;
    end;

// public

    constructor TTokenFile.Create( const aFileName: string); // virtual;
    begin
      inherited Create;
      FLineStack   := TStringList.Create;
      FFileName    := aFileName;
      FbDecimation := 100;
      if ( aFileName <> '') and ( aFileName[ 1] <> '+')
      then LoadFromFile( FFileName);
      Reset;
    end;

    destructor  TTokenFile.Destroy; // override;
    begin
      if Assigned( FOnFileLine)
      then FOnFileLine( Self, FLineNr, FLineNr - FOldLineNr);
      FLineStack.Free;
      inherited Destroy;
    end;

    procedure   TTokenFile.BreakOut; // virtual;
    begin
      // For kids only.
    end;

    function    TTokenFile.GetString( aTerminator: Char): string;

    // done: bug in string parser, " \" did not terminate scanning.

    //
    //   Uses escape characters in tokens
    //
    //   Process '\' characters in a C like manner :
    //
    //     \&nnn      - treat as a decimal       - 3 digits must be present
    //     \$xx       - treat as a hex number    - 2 digits must be present
    //     \%bbbbbbbb - treat as a binary number - 8 digits must be present
    //
    //     \0         - NUL
    //     \a         - BEL
    //     \B         - BS
    //     \F         - FF
    //     \N         - LF
    //     \R         - CR
    //     \T         - TAB
    //     \V         - VT
    //     \X         - same as $ - note C uses three digits, we use 2 here.
    //
    //   Specifically :
    //
    //     \\         - treat as a single \ character
    //     \"         - treat as a single " character
    //     \'         - treat as a single ' character

    var
      Done    : Boolean;
      Escaped : Boolean;
      i       : Integer;
      NumBase : Integer;
      N       : string;
    begin
      Result  := '';
      Done    := False;
      Escaped := False;
      i       := 1;
      NumBase := 0;      // Not scanning a number
      N       := '';
      if FBuffer = ''
      then begin
        Result := endoffile;
        Done   := True;
      end;
      while not Done and ( i <= Length( FBuffer))
      do begin
        if Escaped
        then begin
          case UpCase( FBuffer[ i]) of
            '0' .. '9', 'A' .. 'F' :
              begin
                if NumBase = 0
                then begin
                  case UpCase( FBuffer[ i]) of
                    '0' : Result := Result + #$00;  // NUL
                    'A' : Result := Result + #$07;  // BEL
                    'B' : Result := Result + #$08;  // BS
                    'F' : Result := Result + #$0c;  // FF
                    else  Result := Result + FBuffer[ i];
                  end;
                  Escaped := False;
                end
                else begin
                  N := N + FBuffer[ i];
                  if
                    ((NumBase = 16) and ( Length( N) = 2)) or
                    ((NumBase = 10) and ( Length( N) = 3)) or
                    ((NumBase =  2) and ( Length( N) = 8))
                  then begin
                    Result  := Result + Char( StrToUIntBase( N, NumBase) and $ff);
                    Escaped := False;
                    N       := '';
                    NumBase := 0;
                  end;
                end
              end;
            'N' :                                // line feed
              begin
                Result  := Result + #$0a;
                Escaped := False;
              end;
            'R' :                                // carriage return
              begin
                Result  := Result + #$0d;
                Escaped := False;
              end;
            'T' :                                // tab
              begin
                Result  := Result + #$09;
                Escaped := False;
              end;
            'V' :                                // vertical tab
              begin
                Result  := Result + #$0b;
                Escaped := False;
              end;
            'X', '$' :                           // Hex format
              begin
                Numbase := 16;
                N       := '';
              end;
            '&' :                                // decimal format
              begin
                Numbase := 10;
                N       := '';
              end;
            '%' :                                // binary format
              begin
                Numbase := 2;
                N       := '';
              end;
            else begin
              Result  := Result + FBuffer[ i];
              Escaped := False;
            end;
          end;
        end
        else begin
          case FBuffer[ i] of
            '\' : Escaped := True;
            '''', '"' :
              begin
                if FBuffer[ i] = aTerminator
                then Done := True
                else Result := Result + FBuffer[ i];
              end;
            else Result := Result + FBuffer[ i];
          end;
        end;
        Inc( i);
        Inc( FOffset);
      end;
      if i > 1
      then FBuffer := Copy( FBuffer, i, Length( FBuffer));
    end;

    function    TTokenFile.GetToken( aSeparator: Char; Extended: Boolean): string;
    //
    // if Extended then :
    //
    //   if aSeparator = ' ' and the token to scan starts with '<'
    //   then use [ '>'] as the separator and treat everything between '<' and
    //   '>' as a single token - even when spaces are contained within this
    //   string.
    //
    // escape characters in tokens:
    //
    //   if aSeparator = '"' (the double quote character)
    //   then scan for a string in the following way :
    //
    //   Process '\' characters in a C like manner :
    //
    //     \\         - treat as a single '\' character
    //     \"         - treat as a single '"' character
    //     \nnn       - treat as a decimal       - 3 digits must be present
    //     \$xx       - treat as a hex number    - 2 digits must be present
    //     \%bbbbbbbb - treat as a binary number - 8 difits must be present
    //
    //
    var
      i       : Integer;
      Done    : Boolean;
      StopSet : set of AnsiChar;

      procedure LTrimBuffer;
      var
        i : Integer;
      begin
        if aSeparator = ' '
        then begin
          i := 1;
          while ( i <= Length( FBuffer)) and CharInSet( FBuffer[ i], StopSet)
          do begin
            Inc( FOffset);
            Inc( i);
          end;
          if i > 1
          then FBuffer := Copy( FBuffer, i, Length( FBuffer));
        end;
      end;

    begin
      FCurrToken := '';
      if aSeparator = '"'
      then FCurrToken := GetString( '"')
      else if aSeparator = ''''
      then FCurrToken := GetString( '''')
      else begin
        Done := False;
        if aSeparator = ' '
        then StopSet := WhiteSpace
        else StopSet := [ aSeparator];
        LTrimBuffer;

        while not Done
        do begin
          while (( FLineNr < Count) or ( FLineStack.Count > 0)) and ( FBuffer = '')
          do begin
            FillBuffer;
            LTrimBuffer;
          end;
          if FBuffer = ''
          then begin
            FCurrToken := endoffile;
            Done       := True;
          end
          else begin
            i := 1;
            while not Done and ( i <= Length( FBuffer))
            do begin
              if
                Extended            and
                ( i = 1)            and
                ( aSeparator = ' ') and
                ( FBuffer[ i] = '<')
              then begin
                StopSet  := [ '>'];
                FBuffer  := Copy( FBuffer, 2, Length( FBuffer));
                i        := 1;
                Extended := False; // Don't do this twice !!
              end
              else Inc( i);
              Inc( FOffset);
              Done :=
                (( i <= Length( FBuffer)) and CharInSet( FBuffer[ i], StopSet)) or
                ( aSeparator = ' ') and ( i > Length( FBuffer))
                ;
            end;
            if i > 1
            then begin
              FCurrToken := FCurrToken + Copy( FBuffer, 1, i - 1);
              FBuffer    := Copy( FBuffer, i + 1, Length( FBuffer));
            end
            else FBuffer := '';
          end;
        end;
      end;
      Result := FCurrToken;
      FCurrToken := '';
      FLastToken := Result;
    end;

    procedure   TTokenFile.SkipToEol;
    begin
      FCurrToken := '';
      FBuffer    := '';
    end;

    function    TTokenFile.ReadLine: string;
    begin
      FillBuffer;
      Result := FBuffer;
    end;

    function    TTokenFile.GetToEol: string;
    begin
      Result := FBuffer;
      SkipToEol;
    end;

    procedure   TTokenFile.PushBackBuffer;
    begin
      if FBuffer <> ''
      then PushLine( FBuffer);
      FBuffer := '';
    end;

    procedure   TTokenFile.PushLine( const aLine: string);
    var
      S : string;
      i : Integer;
    begin
      if aLine <> ''
      then begin
        S := '';
        for i := Length( aLine) downto 1
        do begin
          case aLine[ i] of
            ^J : ;                         // Skip line feeds
            ^M : begin                     // Handle Carriage returns
                if S <> ''
                then FLineStack.Add( S);
                S := '';
              end;
            else S := aLine[ i] + S;       // Copy all other
          end;
        end;
        if S <> ''
        then FLineStack.Add( S);
      end
      else FLineStack.Add( aLine);
    end;

    function    TTokenFile.PopLine: string;
    begin
      with FlineStack
      do
        if Count > 0
        then begin
          Result := FLineStack[ Count - 1];
          Delete( Count - 1);
        end
        else Result := '';
    end;


{ ========
  TFileStack = class( TList)
  private
    FOnLine       : TOnLine;
    FOnFileEnter  : TOnFileEnter;
    FOnFileLeave  : TOnFileLeave;
    FOnFileResume : TOnFileResume;
    FTotalCount   : Integer;
    FFbDecimation : Integer;
  private
    property    TokenFile      : TTokenFile    read GetTokenFile;
    property    FileCount      : Integer       read GetFileCount;
  private
    property    FileIndexed[ anIndex: Integer]: TTokenFile read GetFileIndexed;
  public
    property    FileName       : string        read GetFileName;                                  // *
    property    LineNr         : Integer       read GetLineNr;                                    // *
    property    Offset         : Integer       read GetOffset;                                    // *
    property    FbDecimation   : Integer       read FFbDecimation  write SetFbDecimation;         // *
  public
    property    OnLine         : TOnLine       read FOnLine        write FOnLine;                 // *
    property    OnFileEnter    : TOnFileEnter  read FOnFileEnter   write FOnFileEnter;            // *
    property    OnFileLeave    : TOnFileLeave  read FOnFileLeave   write FOnFileLeave;            // *
    property    OnFileResume   : TOnFileResume read FOnFileResume  write FOnFileResume;           // *
  private
}

    function    TFileStack.GetTokenFile: TTokenFile;
    begin
      Result := FileIndexed[ Count - 1];
    end;

    function    TFileStack.GetFileName: string;
    {
    var
      p : Integer;
      T : TTokenFile;
    }
    begin
    {
      p      := FileCount - 1;
      Result := '';
      while p >= 0
      do begin
        T := tokenFile;
        if not( ( T.FileName <> '') and ( T.FileName[ 1] = '+'))
        then begin
          Result := T.FileName;
          Break;
        end;
        Dec( p);
      end;
    }
      if FileCount > 0
      then Result := TokenFile.FileName
      else Result := '';
    end;

    function    TFileStack.GetLineNr: Integer;
    var
      p : Integer;
      T : TTokenFile;
    begin
      p      := FileCount - 1;
      Result := 0;
      while p >= 0
      do begin
        T := tokenFile;
        if ( T.FileName <> '') and ( T.FileName[ 1] = '+')
        then Inc( Result, T.LineNr)
        else begin
          Result := T.LineNr;
          Break;
        end;
        Dec( p);
      end;
    end;

    function    TFileStack.GetOffset: Integer;
    begin
      if FileCount > 0
      then Result := TokenFile.Offset
      else Result := 1;
    end;

    procedure   TFileStack.SetFbDecimation( aValue: Integer);
    begin
      FFbDecimation := aValue;
      if FileCount > 0
      then TokenFile.FbDecimation := aValue;
    end;

    procedure   TFileStack.CloseAll;
    begin
      while FileCount > 0
      do Pop;
    end;

    function    TFileStack.GetFileCount: Integer;
    begin
      Result := Count;
    end;

    function    TFileStack.GetFileIndexed( anIndex: Integer): TTokenFile;
    begin
      if ( anIndex >= 0) and ( anIndex < FileCount)
      then Result := TTokenFile( Items[ anIndex])
      else Result := nil;
    end;

// public

    constructor TFileStack.Create;
    begin
      inherited Create;
      FbDecimation := 100;
    end;

    destructor  TFileStack.Destroy; // override;
    begin
      CloseAll;
      inherited Destroy;
    end;

    procedure   TFileStack.BreakOut;
    begin
      if FileCount > 0
      then TokenFile.BreakOut;
    end;

    procedure   TFileStack.Reset;
    begin
      CloseAll;
      FTotalCount := 0;
    end;

    procedure   TFileStack.Push( const aFileName: string);
    begin
      if Assigned( FOnFileEnter) and ( aFileName <> '')
      then FOnFileEnter( Self, aFileName);
      Add( TTokenFile.Create( aFileName));
      with TokenFile
      do begin
        OnFileLine   := DoFileLine;
        FbDecimation := Self.FbDecimation;
      end;
    end;

    procedure   TFileStack.Pop;
    var
      aFile        : TTokenFile;
      anOldFile    : string;
    begin
      aFile        := TokenFile;
      anOldFile    := aFile.FileName;
      Delete( FileCount - 1);
      aFile.Free;
      if ( anOldFile <> '') and ( anOldFile[ 1] <> '+')
      then begin
        if Assigned( FOnFileLeave)
        then FOnFileLeave( Self, anOldFile);
        if Assigned( FOnFileResume)
        then begin
          if FileCount > 0
          then FOnFileResume( Self, TokenFile.FileName);
        end;
      end;
    end;

    function    TFileStack.GetToken( aSeparator: Char; Extended: Boolean; out aFile: string; out aLine, anOffset: Integer): string;
    begin
      if FileCount > 0
      then
        with TokenFile
        do begin
          Result   := GetToken( aSeparator, Extended);
          aFile    := FileName;
          aLine    := LineNr;
          anOffset := Offset;
          if SameText( Result, endoffile)                                           // then we need recursion, count > 0 : that's ok.
          then begin
            Pop;                                                                    // Lessens count by one
            Result := Self.GetToken( aSeparator, Extended, aFile, aLine, anOffset); // Recur
          end;
        end
      else begin
        Result   := endoffile;
        aFile    := FileName;
        aLine    := LineNr;
        anOffset := Offset;
      end;
    end;

    procedure   TFileStack.SkipToEol;
    begin
      if FileCount > 0
      then TokenFile.SkipToEol;
    end;

    function    TFileStack.ReadLine: string;
    begin
      if FileCount > 0
      then Result := TokenFile.ReadLine
      else Result := endoffile;
    end;

    function    TFileStack.GetToEol: string;
    begin
      if FileCount > 0
      then Result := TokenFile.GetToEol
      else Result := endoffile;
    end;

    procedure   TFileStack.PushBackBuffer;
    begin
      if FileCount > 0
      then TokenFile.PushBackBuffer;
    end;

    procedure   TFileStack.PushLine( const aLine: string; const anID: string);
    begin
      PushBackBuffer;
      if FileName = ''                                 // Allow for pushes on an empty file stack
      then Push( '+' + anID)                           // indicate pushed stuff with a leading +
      else Push( '+' + anID + '+' + FileName);
      with TokenFile
      do PushLine( aLine);
    end;

    procedure   TFileStack.PushLines( const aLines: TStrings; const anID: string);
    var
      i : Integer;
    begin
      if Assigned( aLines) and ( aLines.Count > 0)
      then begin
        PushBackBuffer;
        if FileName = ''                               // Allow for pushes on an empty file stack
        then Push( '+' + anID)                         // indicate pushed stuff with a leading + and an ID
        else Push( '+' + anID + '+' + FileName);
        with aLines
        do begin
          for i:= Count - 1 downto 0
          do TokenFile.PushLine( Strings[ i]);
        end;
      end;
    end;

    procedure   TFileStack.DoFileLine( aSender: TObject; aLineNr, aCount: Integer);
    begin
      Inc( FTotalCount, aCount);
      if Assigned( FOnLine)
      then FOnLine( Self, aLineNr, FTotalCount);
    end;

end.
